@@ -37,6 +37,14 @@ extern "C" {
 
 #include <spa/utils/defs.h>
 
+/** \defgroup spa_json SPA JSON Parser
+ */
+
+/**
+ * \addtogroup spa_json
+ * \{
+ */
+
 /* a simple JSON compatible tokenizer */
 struct spa_json {
 	const char *cur;
@@ -77,7 +85,7 @@ static inline int spa_json_next(struct spa_json * iter, const char **value)
 			goto again;
 		case __STRUCT:
 			switch (cur) {
-			case '\t': case ' ': case '\r': case '\n': case ':': case '=': case ',':
+			case '\0': case '\t': case ' ': case '\r': case '\n': case ':': case '=': case ',':
 				continue;
 			case '#':
 				iter->state = __COMMENT;
@@ -124,8 +132,7 @@ static inline int spa_json_next(struct spa_json * iter, const char **value)
 				iter->state = __STRUCT;
 				if (iter->depth > 0)
 					continue;
-				iter->cur++;
-				return iter->cur - *value;
+				return ++iter->cur - *value;
 			case 240 ... 247:
 				utf8_remain++;
 				SPA_FALLTHROUGH;
@@ -308,7 +315,8 @@ static inline int spa_json_parse_string(const char *val, int len, char *result)
 {
 	const char *p;
 	if (!spa_json_is_string(val, len)) {
-		strncpy(result, val, len);
+		if (result != val)
+			strncpy(result, val, len);
 		result += len;
 	} else {
 		for (p = val+1; p < val + len; p++) {
@@ -324,7 +332,18 @@ static inline int spa_json_parse_string(const char *val, int len, char *result)
 					*result++ = '\t';
 				else if (*p == 'f')
 					*result++ = '\f';
-				else
+				else if (*p == 'u') {
+					char *end;
+					uint16_t v = strtol(p+1, &end, 16);
+					if (p+1 == end) {
+						*result++ = *p;
+					} else {
+						p = end-1;
+						if (v > 0xff)
+							*result++ = (v >> 8) & 0xff;
+						*result++ = v & 0xff;
+					}
+				} else
 					*result++ = *p;
 			} else if (*p == '\"') {
 				break;
@@ -340,7 +359,7 @@ static inline int spa_json_get_string(struct spa_json *iter, char *res, int maxl
 {
 	const char *value;
 	int len;
-	if ((len = spa_json_next(iter, &value)) <= 0 || maxlen < len)
+	if ((len = spa_json_next(iter, &value)) <= 0 || maxlen <= len)
 		return -1;
 	return spa_json_parse_string(value, len, res);
 }
@@ -348,6 +367,7 @@ static inline int spa_json_get_string(struct spa_json *iter, char *res, int maxl
 static inline int spa_json_encode_string(char *str, int size, const char *val)
 {
 	int len = 0;
+	static const char hex[] = { "0123456789abcdef" };
 #define __PUT(c) { if (len < size) *str++ = c; len++; }
 	__PUT('"');
 	while (*val) {
@@ -367,11 +387,18 @@ static inline int spa_json_encode_string(char *str, int size, const char *val)
 		case '\f':
 			__PUT('\\'); __PUT('f');
 			break;
+		case '\\':
 		case '"':
-			__PUT('\\'); __PUT('"');
+			__PUT('\\'); __PUT(*val);
 			break;
 		default:
-			__PUT(*val);
+			if (*val > 0 && *val < 0x20) {
+				__PUT('\\'); __PUT('u');
+				__PUT('0'); __PUT('0');
+				__PUT(hex[((*val)>>4)&0xf]); __PUT(hex[(*val)&0xf]);
+			} else {
+				__PUT(*val);
+			}
 			break;
 		}
 		val++;
@@ -382,6 +409,10 @@ static inline int spa_json_encode_string(char *str, int size, const char *val)
 	return len-1;
 }
 
+/**
+ * \}
+ */
+
 #ifdef __cplusplus
 } /* extern "C" */
 #endif
